; main.s
; Runs on Launchpad (TM4c and LMF120)
; A set of subroutines to show parameter passing
; options and considerations for the ARM
; Ramesh Yerraballi
; Feb 10, 2013
; Lessons:
;  a. Parameter passing using registers should always be the first option
;  b. To make your code readable use symbolic names for registers (with the RN directive)
;     where you refer to a register by what its holding and hence being human readable
;  c. To the extent possible write assembly subroutines so the input parameters are 
;     assumed to be in registers R0, R1, R2 and R4; This will guarantee a long
;     life for your assembly routines because they will be compatible with 
;     C programs that call this subroutine. For example the Exp subroutine (below)
;     can be called from C as follows: answer = Exp(2,3); 
;  d. Use parameter-passing through stack when passing more than 4 parameters


       THUMB
;  RAM Section
       AREA    DATA, ALIGN=2
	   ; These variables are not used in this program 
med   SPACE   2    
big   SPACE   4
small SPACE   1
       ALIGN
;  EEPROM Code Section	   
       AREA    |.text|, CODE, READONLY, ALIGN=2
       EXPORT  Start
Start  
	MOV	R0,#7
	BL  Square
   ;; R1 has the result 49
 ;-------- call a subroutine thay uses resgisters for parameter passing
 ;  Not AAPCS compliant
	MOV R1,#3     ; Assume R0 from previous is first input and set R1 to 3
	BL  Exp
   ;; R2 has the result 7^3 = 343 (0x157) 
 ;-------- call a subroutine that uses stack for parameter passing
 ;  Not AAPCS compliant
    MOV  R0,#12
	MOV  R1,#5
	MOV  R2,#22
	MOV  R3,#7
	MOV  R4,#18
	PUSH {R0-R4}	; Stack has 12,5,22,7 and 18 (with 12 on top)
	BL  Max5		; Call Max5 to find the maximum of the five numbers
	POP {R5}        
	;; R5 has the max element (22) 
;--------- call a subroutine that uses both stack and registers for parameter passing: 
;  AAPCS compliant on inputs but fails to comply on return (two outputs instead of 1)	
    MOV  R0,#6	;Element count
	MOV  R1,#-14
	MOV  R2,#5
	MOV  R3,#0 
	MOV  R4,#-7
	MOV  R5,#32
	MOV  R6,#-5
	PUSH {R4-R6}	; R0 has element count, R1-R3 have first 3 inputs and Stack has the remaining
	BL  MinMax
	;; R0 has -14 and R1 has 32 upon return
    SWI 0	; Terminate program	

;------------Square------------
; Input: R0  has input number (num)
; Output: R1 has the output which is the square of the input
; Comments: Input and output are signed.
;           Not AAPCS compliant (uses R1 to return single result instead of R0)
num		RN	0		; The RN directive makes it easy to give names
out		RN	1		; to registers that make subroutine inputs more human readable
Square 
	   MUL  out,num,num
	BX  LR

;---------Exp-----------
; Input: R0 and R1 have inputs XX an YY
; Output: R2 has the result XX raised to the power YY
; Comments: R1 and R2 and non-negative
;           Destroys input R1
;           Not AAPCS compliant (uses R3 to return single result instead of R0)
XX		RN	0		; The RN directive makes it easy to give names
YY		RN	1		; to registers that make subroutine inputs more human readable
pow     RN  2
Exp 
	   ADDS XX,#0		
	   BEQ	Zero
	   ADDS	YY,#0		; check if YY is zero
	   BEQ  One
	   MOV  pow, #1		; Initial product is 1 
More   MUL  pow,XX      ; Add to XX to partial product
	   ADDS YY,#-1		; Decrement YY
	   BNE	More
	   B    Retn        ; Done adding so return
Zero	   
	   MOV	pow,#0		; XX is 0 so result is 0
       B    Retn
One    MOV  pow,#1
Retn   BX  LR

;---------Max5-----------
; Input: 5 numbers pushed on the stack
; Output: put only the maximum number on the stack
; Comments: The input numbers are removed from stack
numM	RN	1	; hold the current number in numM
max     RN  2   ; hold maximum so far in max
count   RN  0   ; how many elements
Max5
   
	   POP {max}		; get top element (top of stack)
	   	            ; store it in max
	   MOV count,#4 ; 4 more to go
Again  POP {numM}		; get next element
	   CMP numM,max
	   BLT Next
	   MOV max, numM   ; new numM is the max
Next   ADDS count,#-1  ; one more checked
	   BNE Again
	   PUSH {max}       ; found max so push it on stack
	   BX  LR
	   
;---------MinMax-----------
; Input: N numbers pushed on the stack
;        N passed in R0
; Output: Return in R0 the min and R1 the max 
; Comments: The input numbers are removed from stack
;           Destroys input R0 (N)
numMM	RN	3		; hold the current number in number
maxMM     RN  1     ; hold maximum so far in max; can't call it 
                    ; max as that name is already used to mean R2 (above)
minMM     RN  2
N       RN  0		; how many elements
MinMax
	   PUSH {R1-R3}  ; All elements on Stack
	   CMP N,#0		; if N is zero nothing to do
	   BEQ   DoneMM
	   POP {minMM}	; pop top and set it
	   MOV maxMM,minMM  ;  as the current min and max
loop   ADDS N,#-1    ;  decrement and check
	   BEQ  DoneMM
	   POP {numMM}
	   CMP numMM,maxMM
	   BLT Chkmin	
	   MOV maxMM,numMM  ; new num is the max
Chkmin CMP numMM,minMM
       BGT NextMM     
	   MOV minMM,numMM  ; new num is the min
NextMM B   loop
DoneMM MOV R0,minMM	  ; R0 must have min
	   BX  LR 
       ALIGN      
       END  
           